@@ -4,12 +4,14 @@
 
 use crate::prelude::*;
 
+use rand::distr::SampleString as _;
+use sha2::Digest as _;
+use tokio::io::AsyncReadExt as _;
+
 use crate::json::{
     DeserializeJsonWithPath as _, DeserializeJsonWithPathAsync as _,
 };
 
-use tokio::io::AsyncReadExt as _;
-
 #[derive(
     serde_repr::Serialize_repr,
     serde_repr::Deserialize_repr,
@@ -58,7 +60,6 @@ pub enum TwoFactorProviderType {
 }
 
 impl TwoFactorProviderType {
-    #[must_use]
     pub fn message(&self) -> &str {
         match *self {
             Self::Authenticator => "Enter the 6 digit verification code from your authenticator app.",
@@ -68,7 +69,6 @@ impl TwoFactorProviderType {
         }
     }
 
-    #[must_use]
     pub fn header(&self) -> &str {
         match *self {
             Self::Authenticator => "Authenticator App",
@@ -78,7 +78,6 @@ impl TwoFactorProviderType {
         }
     }
 
-    #[must_use]
     pub fn grab(&self) -> bool {
         !matches!(self, Self::Email)
     }
@@ -90,7 +89,7 @@ impl<'de> serde::Deserialize<'de> for TwoFactorProviderType {
         D: serde::Deserializer<'de>,
     {
         struct TwoFactorProviderTypeVisitor;
-        impl<'de> serde::de::Visitor<'de> for TwoFactorProviderTypeVisitor {
+        impl serde::de::Visitor<'_> for TwoFactorProviderTypeVisitor {
             type Value = TwoFactorProviderType;
 
             fn expecting(
@@ -176,7 +175,7 @@ impl<'de> serde::Deserialize<'de> for KdfType {
         D: serde::Deserializer<'de>,
     {
         struct KdfTypeVisitor;
-        impl<'de> serde::de::Visitor<'de> for KdfTypeVisitor {
+        impl serde::de::Visitor<'_> for KdfTypeVisitor {
             type Value = KdfType;
 
             fn expecting(
@@ -254,6 +253,21 @@ impl serde::Serialize for KdfType {
     }
 }
 
+#[derive(
+    serde_repr::Serialize_repr,
+    serde_repr::Deserialize_repr,
+    Debug,
+    Copy,
+    Clone,
+    PartialEq,
+    Eq,
+)]
+#[repr(u8)]
+pub enum CipherRepromptType {
+    None = 0,
+    Password = 1,
+}
+
 #[derive(serde::Serialize, Debug)]
 struct PreloginReq {
     email: String,
@@ -272,13 +286,10 @@ struct PreloginRes {
 }
 
 #[derive(serde::Serialize, Debug)]
-struct ConnectPasswordReq {
+struct ConnectTokenReq {
     grant_type: String,
-    username: String,
-    password: Option<String>,
     scope: String,
     client_id: String,
-    client_secret: Option<String>,
     #[serde(rename = "deviceType")]
     device_type: u32,
     #[serde(rename = "deviceIdentifier")]
@@ -291,10 +302,39 @@ struct ConnectPasswordReq {
     two_factor_token: Option<String>,
     #[serde(rename = "twoFactorProvider")]
     two_factor_provider: Option<u32>,
+    #[serde(flatten)]
+    auth: ConnectTokenAuth,
+}
+
+#[derive(serde::Serialize, Debug)]
+#[serde(untagged)]
+enum ConnectTokenAuth {
+    Password(ConnectTokenPassword),
+    AuthCode(ConnectTokenAuthCode),
+    ClientCredentials(ConnectTokenClientCredentials),
+}
+
+#[derive(serde::Serialize, Debug)]
+struct ConnectTokenPassword {
+    username: String,
+    password: String,
+}
+
+#[derive(serde::Serialize, Debug)]
+struct ConnectTokenAuthCode {
+    code: String,
+    code_verifier: String,
+    redirect_uri: String,
+}
+
+#[derive(serde::Serialize, Debug)]
+struct ConnectTokenClientCredentials {
+    username: String,
+    client_secret: String,
 }
 
 #[derive(serde::Deserialize, Debug)]
-struct ConnectPasswordRes {
+struct ConnectTokenRes {
     access_token: String,
     refresh_token: String,
     #[serde(rename = "Key", alias = "key")]
@@ -309,6 +349,11 @@ struct ConnectErrorRes {
     error_model: Option<ConnectErrorResErrorModel>,
     #[serde(rename = "TwoFactorProviders", alias = "twoFactorProviders")]
     two_factor_providers: Option<Vec<TwoFactorProviderType>>,
+    #[serde(
+        rename = "SsoEmail2faSessionToken",
+        alias = "ssoEmail2faSessionToken"
+    )]
+    sso_email_2fa_session_token: Option<String>,
 }
 
 #[derive(serde::Deserialize, Debug)]
@@ -329,6 +374,18 @@ struct ConnectRefreshTokenRes {
     access_token: String,
 }
 
+#[derive(serde::Serialize, Debug)]
+struct SendEmailLoginReq {
+    email: String,
+    #[serde(rename = "DeviceIdentifier", alias = "deviceIdentifier")]
+    device_identifier: String,
+    #[serde(
+        rename = "SsoEmail2faSessionToken",
+        alias = "ssoEmail2faSessionToken"
+    )]
+    sso_email_2fa_session_token: String,
+}
+
 #[derive(serde::Deserialize, Debug)]
 struct SyncRes {
     #[serde(rename = "Ciphers", alias = "ciphers")]
@@ -357,6 +414,8 @@ struct SyncResCipher {
     identity: Option<CipherIdentity>,
     #[serde(rename = "SecureNote", alias = "secureNote")]
     secure_note: Option<CipherSecureNote>,
+    #[serde(rename = "SshKey", alias = "sshKey")]
+    ssh_key: Option<CipherSshKey>,
     #[serde(rename = "Notes", alias = "notes")]
     notes: Option<String>,
     #[serde(rename = "PasswordHistory", alias = "passwordHistory")]
@@ -365,6 +424,10 @@ struct SyncResCipher {
     fields: Option<Vec<CipherField>>,
     #[serde(rename = "DeletedDate", alias = "deletedDate")]
     deleted_date: Option<String>,
+    #[serde(rename = "Key", alias = "key")]
+    key: Option<String>,
+    #[serde(rename = "Reprompt", alias = "reprompt")]
+    reprompt: CipherRepromptType,
 }
 
 impl SyncResCipher {
@@ -456,6 +519,12 @@ impl SyncResCipher {
             }
         } else if let Some(_secure_note) = &self.secure_note {
             crate::db::EntryData::SecureNote
+        } else if let Some(ssh_key) = &self.ssh_key {
+            crate::db::EntryData::SshKey {
+                private_key: ssh_key.private_key.clone(),
+                public_key: ssh_key.public_key.clone(),
+                fingerprint: ssh_key.fingerprint.clone(),
+            }
         } else {
             return None;
         };
@@ -480,6 +549,8 @@ impl SyncResCipher {
             fields,
             notes: self.notes.clone(),
             history,
+            key: self.key.clone(),
+            master_password_reprompt: self.reprompt,
         })
     }
 }
@@ -532,7 +603,7 @@ struct CipherLoginUri {
 
 #[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
 struct CipherCard {
-    #[serde(rename = "CardholderName", alias = "cardHolderName")]
+    #[serde(rename = "CardholderName", alias = "cardholderName")]
     cardholder_name: Option<String>,
     #[serde(rename = "Number", alias = "number")]
     number: Option<String>,
@@ -584,6 +655,16 @@ struct CipherIdentity {
     username: Option<String>,
 }
 
+#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
+struct CipherSshKey {
+    #[serde(rename = "PrivateKey", alias = "privateKey")]
+    private_key: Option<String>,
+    #[serde(rename = "PublicKey", alias = "publicKey")]
+    public_key: Option<String>,
+    #[serde(rename = "Fingerprint", alias = "keyFingerprint")]
+    fingerprint: Option<String>,
+}
+
 #[derive(
     serde_repr::Serialize_repr,
     serde_repr::Deserialize_repr,
@@ -644,7 +725,7 @@ pub enum LinkedIdType {
 #[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
 struct CipherField {
     #[serde(rename = "Type", alias = "type")]
-    ty: FieldType,
+    ty: Option<FieldType>,
     #[serde(rename = "Name", alias = "name")]
     name: Option<String>,
     #[serde(rename = "Value", alias = "value")]
@@ -701,12 +782,6 @@ struct CiphersPutReq {
     password_history: Vec<CiphersPutReqHistory>,
 }
 
-#[derive(serde::Serialize, Debug)]
-struct CiphersPutReqLogin {
-    username: Option<String>,
-    password: Option<String>,
-}
-
 #[derive(serde::Serialize, Debug)]
 struct CiphersPutReqHistory {
     #[serde(rename = "LastUsedDate")]
@@ -734,29 +809,59 @@ struct FoldersPostReq {
     name: String,
 }
 
+// Used for the Bitwarden-Client-Name header. Accepted values:
+// https://github.com/bitwarden/server/blob/main/src/Core/Enums/BitwardenClient.cs
+const BITWARDEN_CLIENT: &str = "cli";
+
+// DeviceType.LinuxDesktop, as per Bitwarden API device types.
+const DEVICE_TYPE: u8 = 8;
+
 #[derive(Debug)]
 pub struct Client {
     base_url: String,
     identity_url: String,
+    ui_url: String,
     client_cert_path: Option<std::path::PathBuf>,
 }
 
 impl Client {
-    #[must_use]
     pub fn new(
         base_url: &str,
         identity_url: &str,
+        ui_url: &str,
         client_cert_path: Option<&std::path::Path>,
     ) -> Self {
         Self {
             base_url: base_url.to_string(),
             identity_url: identity_url.to_string(),
+            ui_url: ui_url.to_string(),
             client_cert_path: client_cert_path
                 .map(std::path::Path::to_path_buf),
         }
     }
 
     async fn reqwest_client(&self) -> Result<reqwest::Client> {
+        let mut default_headers = axum::http::HeaderMap::new();
+        default_headers.insert(
+            "Bitwarden-Client-Name",
+            axum::http::HeaderValue::from_static(BITWARDEN_CLIENT),
+        );
+        default_headers.insert(
+            "Bitwarden-Client-Version",
+            axum::http::HeaderValue::from_static(env!("CARGO_PKG_VERSION")),
+        );
+        default_headers.append(
+            "Device-Type",
+            // unwrap is safe here because DEVICE_TYPE is a number and digits
+            // are valid ASCII
+            axum::http::HeaderValue::from_str(&DEVICE_TYPE.to_string())
+                .unwrap(),
+        );
+        let user_agent = format!(
+            "{}/{}",
+            env!("CARGO_PKG_NAME"),
+            env!("CARGO_PKG_VERSION")
+        );
         if let Some(client_cert_path) = self.client_cert_path.as_ref() {
             let mut buf = Vec::new();
             let mut f = tokio::fs::File::open(client_cert_path)
@@ -774,21 +879,15 @@ impl Client {
             let pem = reqwest::Identity::from_pem(&buf)
                 .map_err(|e| Error::CreateReqwestClient { source: e })?;
             Ok(reqwest::Client::builder()
-                .user_agent(format!(
-                    "{}/{}",
-                    env!("CARGO_PKG_NAME"),
-                    env!("CARGO_PKG_VERSION")
-                ))
+                .user_agent(user_agent)
                 .identity(pem)
+                .default_headers(default_headers)
                 .build()
                 .map_err(|e| Error::CreateReqwestClient { source: e })?)
         } else {
             Ok(reqwest::Client::builder()
-                .user_agent(format!(
-                    "{}/{}",
-                    env!("CARGO_PKG_NAME"),
-                    env!("CARGO_PKG_VERSION")
-                ))
+                .user_agent(user_agent)
+                .default_headers(default_headers)
                 .build()
                 .map_err(|e| Error::CreateReqwestClient { source: e })?)
         }
@@ -803,7 +902,7 @@ impl Client {
         };
         let client = self.reqwest_client().await?;
         let res = client
-            .post(&self.api_url("/accounts/prelogin"))
+            .post(self.identity_url("/accounts/prelogin"))
             .json(&prelogin)
             .send()
             .await
@@ -823,18 +922,22 @@ impl Client {
         device_id: &str,
         apikey: &crate::locked::ApiKey,
     ) -> Result<()> {
-        let connect_req = ConnectPasswordReq {
+        let connect_req = ConnectTokenReq {
+            auth: ConnectTokenAuth::ClientCredentials(
+                ConnectTokenClientCredentials {
+                    username: email.to_string(),
+                    client_secret: String::from_utf8(
+                        apikey.client_secret().to_vec(),
+                    )
+                    .unwrap(),
+                },
+            ),
             grant_type: "client_credentials".to_string(),
-            username: email.to_string(),
-            password: None,
             scope: "api".to_string(),
             // XXX unwraps here are not necessarily safe
             client_id: String::from_utf8(apikey.client_id().to_vec())
                 .unwrap(),
-            client_secret: Some(
-                String::from_utf8(apikey.client_secret().to_vec()).unwrap(),
-            ),
-            device_type: 8,
+            device_type: u32::from(DEVICE_TYPE),
             device_identifier: device_id.to_string(),
             device_name: "rbw".to_string(),
             device_push_token: String::new(),
@@ -843,7 +946,7 @@ impl Client {
         };
         let client = self.reqwest_client().await?;
         let res = client
-            .post(&self.identity_url("/connect/token"))
+            .post(self.identity_url("/connect/token"))
             .form(&connect_req)
             .send()
             .await
@@ -871,29 +974,58 @@ impl Client {
     pub async fn login(
         &self,
         email: &str,
+        sso_id: Option<&str>,
         device_id: &str,
         password_hash: &crate::locked::PasswordHash,
         two_factor_token: Option<&str>,
         two_factor_provider: Option<TwoFactorProviderType>,
     ) -> Result<(String, String, String)> {
-        let connect_req = ConnectPasswordReq {
-            grant_type: "password".to_string(),
-            username: email.to_string(),
-            password: Some(crate::base64::encode(password_hash.hash())),
-            scope: "api offline_access".to_string(),
-            client_id: "desktop".to_string(),
-            client_secret: None,
-            device_type: 8,
-            device_identifier: device_id.to_string(),
-            device_name: "rbw".to_string(),
-            device_push_token: String::new(),
-            two_factor_token: two_factor_token
-                .map(std::string::ToString::to_string),
-            two_factor_provider: two_factor_provider.map(|ty| ty as u32),
+        let connect_req = match sso_id {
+            Some(sso_id) => {
+                let (sso_code, sso_code_verifier, callback_url) =
+                    self.obtain_sso_code(sso_id).await?;
+
+                ConnectTokenReq {
+                    auth: ConnectTokenAuth::AuthCode(ConnectTokenAuthCode {
+                        code: sso_code,
+                        code_verifier: sso_code_verifier,
+                        redirect_uri: callback_url,
+                    }),
+                    grant_type: "authorization_code".to_string(),
+                    scope: "api offline_access".to_string(),
+                    client_id: "cli".to_string(),
+                    device_type: u32::from(DEVICE_TYPE),
+                    device_identifier: device_id.to_string(),
+                    device_name: "rbw".to_string(),
+                    device_push_token: String::new(),
+                    two_factor_token: two_factor_token
+                        .map(std::string::ToString::to_string),
+                    two_factor_provider: two_factor_provider
+                        .map(|ty| ty as u32),
+                }
+            }
+            None => ConnectTokenReq {
+                auth: ConnectTokenAuth::Password(ConnectTokenPassword {
+                    username: email.to_string(),
+                    password: crate::base64::encode(password_hash.hash()),
+                }),
+
+                grant_type: "password".to_string(),
+                scope: "api offline_access".to_string(),
+                client_id: "cli".to_string(),
+                device_type: 8,
+                device_identifier: device_id.to_string(),
+                device_name: "rbw".to_string(),
+                device_push_token: String::new(),
+                two_factor_token: two_factor_token
+                    .map(std::string::ToString::to_string),
+                two_factor_provider: two_factor_provider.map(|ty| ty as u32),
+            },
         };
+
         let client = self.reqwest_client().await?;
         let res = client
-            .post(&self.identity_url("/connect/token"))
+            .post(self.identity_url("/connect/token"))
             .form(&connect_req)
             .header(
                 "auth-email",
@@ -902,9 +1034,9 @@ impl Client {
             .send()
             .await
             .map_err(|source| Error::Reqwest { source })?;
+
         if res.status() == reqwest::StatusCode::OK {
-            let connect_res: ConnectPasswordRes =
-                res.json_with_path().await?;
+            let connect_res: ConnectTokenRes = res.json_with_path().await?;
             Ok((
                 connect_res.access_token,
                 connect_res.refresh_token,
@@ -928,6 +1060,91 @@ impl Client {
         }
     }
 
+    pub async fn send_email_login(
+        &self,
+        email: &str,
+        device_id: &str,
+        sso_email_2fa_session_token: &str,
+    ) -> Result<()> {
+        let send_email_login_req = SendEmailLoginReq {
+            email: email.to_string(),
+            device_identifier: device_id.to_string(),
+            sso_email_2fa_session_token: sso_email_2fa_session_token
+                .to_string(),
+        };
+
+        let client = self.reqwest_client().await?;
+        let res = client
+            .post(self.api_url("/two-factor/send-email-login"))
+            .json(&send_email_login_req)
+            .header(
+                "auth-email",
+                crate::base64::encode_url_safe_no_pad(email),
+            )
+            .send()
+            .await
+            .map_err(|source| Error::Reqwest { source })?;
+
+        if res.status() == reqwest::StatusCode::OK {
+            Ok(())
+        } else {
+            let code = res.status().as_u16();
+            log::warn!("{code}: {:?}", res.text().await);
+            Err(Error::RequestFailed { status: code })
+        }
+    }
+
+    async fn obtain_sso_code(
+        &self,
+        sso_id: &str,
+    ) -> Result<(String, String, String)> {
+        let state =
+            rand::distr::Alphanumeric.sample_string(&mut rand::rng(), 64);
+        let sso_code_verifier =
+            rand::distr::Alphanumeric.sample_string(&mut rand::rng(), 64);
+
+        let mut hasher = sha2::Sha256::new();
+        hasher.update(sso_code_verifier.clone());
+        let code_challenge =
+            crate::base64::encode_url_safe_no_pad(hasher.finalize());
+
+        let port = find_free_port(8065, 8070).await?;
+
+        let listener = tokio::net::TcpListener::bind(("127.0.0.1", port))
+            .await
+            .map_err(|e| Error::CreateSSOCallbackServer { err: e })?;
+
+        let callback_server =
+            start_sso_callback_server(listener, state.as_str());
+
+        let callback_url =
+            "http://localhost:".to_string() + port.to_string().as_str();
+
+        open::that(
+            self.ui_url.clone()
+                + "/#/sso?clientId="
+                + "cli"
+                + "&redirectUri="
+                + urlencoding::encode(callback_url.as_str())
+                    .into_owned()
+                    .as_str()
+                + "&state="
+                + state.as_str()
+                + "&codeChallenge="
+                + code_challenge.as_str()
+                + "&identifier="
+                + sso_id,
+        )
+        .map_err(|e| Error::FailedToOpenWebBrowser { err: e })?;
+        // TODO: probably it'd be better to display the URL in the console if the automatic
+        // open operation fails, instead of failing the whole process? E.g. docker container
+        // case
+
+        let sso_code = callback_server.await?;
+
+        Ok((sso_code, sso_code_verifier, callback_url))
+    }
+
     pub async fn sync(
         &self,
         access_token: &str,
@@ -939,8 +1156,10 @@ impl Client {
     )> {
         let client = self.reqwest_client().await?;
         let res = client
-            .get(&self.api_url("/sync"))
+            .get(self.api_url("/sync"))
             .header("Authorization", format!("Bearer {access_token}"))
+            // This is necessary for vaultwarden to include the ssh keys in the response
+            .header("Bitwarden-Client-Version", "2024.12.0")
             .send()
             .await
             .map_err(|source| Error::Reqwest { source })?;
@@ -1006,7 +1225,7 @@ impl Client {
                     Some(
                         uris.iter()
                             .map(|s| CipherLoginUri {
-                                uri: Some(s.uri.to_string()),
+                                uri: Some(s.uri.clone()),
                                 match_type: s.match_type,
                             })
                             .collect(),
@@ -1075,9 +1294,10 @@ impl Client {
                     username: username.clone(),
                 });
             }
-            crate::db::EntryData::SecureNote {} => {
+            crate::db::EntryData::SecureNote => {
                 req.secure_note = Some(CipherSecureNote {});
             }
+            crate::db::EntryData::SshKey { .. } => unreachable!(),
         }
         let client = reqwest::blocking::Client::new();
         let res = client
@@ -1112,9 +1332,10 @@ impl Client {
         let mut req = CiphersPutReq {
             ty: match data {
                 crate::db::EntryData::Login { .. } => 1,
-                crate::db::EntryData::SecureNote { .. } => 2,
+                crate::db::EntryData::SecureNote => 2,
                 crate::db::EntryData::Card { .. } => 3,
                 crate::db::EntryData::Identity { .. } => 4,
+                crate::db::EntryData::SshKey { .. } => unreachable!(),
             },
             folder_id: folder_uuid.map(std::string::ToString::to_string),
             organization_id: org_id.map(std::string::ToString::to_string),
@@ -1154,7 +1375,7 @@ impl Client {
                     Some(
                         uris.iter()
                             .map(|s| CipherLoginUri {
-                                uri: Some(s.uri.to_string()),
+                                uri: Some(s.uri.clone()),
                                 match_type: s.match_type,
                             })
                             .collect(),
@@ -1223,9 +1444,10 @@ impl Client {
                     username: username.clone(),
                 });
             }
-            crate::db::EntryData::SecureNote {} => {
+            crate::db::EntryData::SecureNote => {
                 req.secure_note = Some(CipherSecureNote {});
             }
+            crate::db::EntryData::SshKey { .. } => unreachable!(),
         }
         let client = reqwest::blocking::Client::new();
         let res = client
@@ -1326,7 +1548,7 @@ impl Client {
     ) -> Result<String> {
         let connect_req = ConnectRefreshTokenReq {
             grant_type: "refresh_token".to_string(),
-            client_id: "desktop".to_string(),
+            client_id: "cli".to_string(),
             refresh_token: refresh_token.to_string(),
         };
         let client = reqwest::blocking::Client::new();
@@ -1345,12 +1567,12 @@ impl Client {
     ) -> Result<String> {
         let connect_req = ConnectRefreshTokenReq {
             grant_type: "refresh_token".to_string(),
-            client_id: "desktop".to_string(),
+            client_id: "cli".to_string(),
             refresh_token: refresh_token.to_string(),
         };
         let client = self.reqwest_client().await?;
         let res = client
-            .post(&self.identity_url("/connect/token"))
+            .post(self.identity_url("/connect/token"))
             .form(&connect_req)
             .send()
             .await
@@ -1369,6 +1591,122 @@ impl Client {
     }
 }
 
+async fn find_free_port(bottom: u16, top: u16) -> Result<u16> {
+    for port in bottom..top {
+        if tokio::net::TcpListener::bind(("127.0.0.1", port))
+            .await
+            .is_ok()
+        {
+            return Ok(port);
+        }
+    }
+
+    Err(Error::FailedToFindFreePort {
+        range: format!("({bottom}..{top})"),
+    })
+}
+
+#[derive(Clone)]
+struct SSOHandlerState {
+    state: String,
+    sender: tokio::sync::mpsc::Sender<Result<String>>,
+}
+
+async fn start_sso_callback_server(
+    listener: tokio::net::TcpListener,
+    state: &str,
+) -> Result<String> {
+    let (shut_sender, shut_receiver) = tokio::sync::mpsc::channel(1);
+    let (sender, mut receiver) = tokio::sync::mpsc::channel(1);
+
+    let sso_handler_state = std::sync::Arc::new(SSOHandlerState {
+        state: state.to_string(),
+        sender: shut_sender,
+    });
+
+    let app = axum::Router::new()
+        .route("/", axum::routing::get(handle_sso_callback))
+        .with_state(sso_handler_state);
+
+    axum::serve(listener, app)
+        .with_graceful_shutdown(sso_server_graceful_shutdown(
+            sender,
+            shut_receiver,
+        ))
+        .await
+        .map_err(|e| Error::FailedToProcessSSOCallback {
+            msg: e.to_string(),
+        })?;
+
+    receiver.recv().await.unwrap()
+}
+
+async fn sso_server_graceful_shutdown(
+    sender: tokio::sync::mpsc::Sender<Result<String>>,
+    mut receiver: tokio::sync::mpsc::Receiver<Result<String>>,
+) {
+    sender.send(receiver.recv().await.unwrap()).await.unwrap();
+}
+
+async fn handle_sso_callback(
+    axum::extract::State(state): axum::extract::State<
+        std::sync::Arc<SSOHandlerState>,
+    >,
+    axum::extract::Query(params): axum::extract::Query<
+        std::collections::HashMap<String, String>,
+    >,
+) -> axum::http::Response<String> {
+    match sso_query_code(&params, state.state.as_str()) {
+        Ok(sso_code) => {
+            state.sender.send(Ok(sso_code)).await.unwrap();
+
+            axum::http::Response::builder().status(axum::http::StatusCode::OK).
+            body(
+                "<html><head><title>Success | rbw</title></head><body> \
+                  <h1>Successfully authenticated with rbw</h1> \
+                  <p>You may now close this tab and return to the terminal.</p> \
+                  </body></html>".to_string()).unwrap()
+        }
+        Err(e) => {
+            state.sender.send(Err(e)).await.unwrap();
+
+            axum::http::Response::builder().status(axum::http::StatusCode::BAD_REQUEST).
+            body(
+                "<html><head><title>Failed | rbw</title></head><body> \
+                  <h1>Something went wrong logging into the rbw</h1> \
+                  <p>You may now close this tab and return to the terminal.</p> \
+                  </body></html>".to_string()).unwrap()
+        }
+    }
+}
+
+fn sso_query_code(
+    params: &std::collections::HashMap<String, String>,
+    state: &str,
+) -> Result<String> {
+    let sso_code =
+        params
+            .get("code")
+            .ok_or(Error::FailedToProcessSSOCallback {
+                msg: "Could not obtain code from the URL".to_string(),
+            })?;
+
+    let received_state =
+        params
+            .get("state")
+            .ok_or(Error::FailedToProcessSSOCallback {
+                msg: "Could not obtain state from the URL".to_string(),
+            })?;
+
+    if received_state.split("_identifier=").next().unwrap() != state {
+        return Err(Error::FailedToProcessSSOCallback {
+            msg: format!("SSO callback states do not match, sent: {state}, received: {received_state}"),
+        });
+    }
+
+    Ok(sso_code.clone())
+}
+
 fn classify_login_error(error_res: &ConnectErrorRes, code: u16) -> Error {
     let error_desc = error_res.error_description.clone();
     let error_desc = error_desc.as_deref();
@@ -1386,6 +1724,9 @@ fn classify_login_error(error_res: &ConnectErrorRes, code: u16) -> Error {
                 {
                     return Error::TwoFactorRequired {
                         providers: providers.clone(),
+                        sso_email_2fa_session_token: error_res
+                            .sso_email_2fa_session_token
+                            .clone(),
                     };
                 }
             }
@@ -1422,6 +1763,6 @@ fn classify_login_error(error_res: &ConnectErrorRes, code: u16) -> Error {
         _ => {}
     }
 
-    log::warn!("unexpected error received during login: {:?}", error_res);
+    log::warn!("unexpected error received during login: {error_res:?}");
     Error::RequestFailed { status: code }
 }
