@@ -1,8 +1,9 @@
 use anyhow::Context as _;
+use sha2::Digest as _;
 
 pub async fn register(
     sock: &mut crate::sock::Sock,
-    tty: Option<&str>,
+    environment: &rbw::protocol::Environment,
 ) -> anyhow::Result<()> {
     let db = load_db().await.unwrap_or_else(|_| rbw::db::Db::new());
 
@@ -12,8 +13,7 @@ pub async fn register(
             .context("failed to parse base url")?;
         let Some(host) = url.host_str() else {
             return Err(anyhow::anyhow!(
-                "couldn't find host in rbw base url {}",
-                url_str
+                "couldn't find host in rbw base url {url_str}"
             ));
         };
 
@@ -33,7 +33,7 @@ pub async fn register(
                 "API key client__id",
                 &format!("Log in to {host}"),
                 err.as_deref(),
-                tty,
+                environment,
                 false,
             )
             .await
@@ -43,7 +43,7 @@ pub async fn register(
                 "API key client__secret",
                 &format!("Log in to {host}"),
                 err.as_deref(),
-                tty,
+                environment,
                 false,
             )
             .await
@@ -61,7 +61,6 @@ pub async fn register(
                         .context("failed to log in to bitwarden instance");
                     }
                     err_msg = Some(message);
-                    continue;
                 }
                 Err(e) => {
                     return Err(e)
@@ -78,8 +77,8 @@ pub async fn register(
 
 pub async fn login(
     sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
-    tty: Option<&str>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    environment: &rbw::protocol::Environment,
 ) -> anyhow::Result<()> {
     let db = load_db().await.unwrap_or_else(|_| rbw::db::Db::new());
 
@@ -89,8 +88,7 @@ pub async fn login(
             .context("failed to parse base url")?;
         let Some(host) = url.host_str() else {
             return Err(anyhow::anyhow!(
-                "couldn't find host in rbw base url {}",
-                url_str
+                "couldn't find host in rbw base url {url_str}"
             ));
         };
 
@@ -110,7 +108,7 @@ pub async fn login(
                 "Master Password",
                 &format!("Log in to {host}"),
                 err.as_deref(),
-                tty,
+                environment,
                 true,
             )
             .await
@@ -143,7 +141,10 @@ pub async fn login(
                     .await?;
                     break 'attempts;
                 }
-                Err(rbw::error::Error::TwoFactorRequired { providers }) => {
+                Err(rbw::error::Error::TwoFactorRequired {
+                    providers,
+                    sso_email_2fa_session_token,
+                }) => {
                     let supported_types = vec![
                         rbw::api::TwoFactorProviderType::Authenticator,
                         rbw::api::TwoFactorProviderType::Yubikey,
@@ -152,6 +153,19 @@ pub async fn login(
 
                     for provider in supported_types {
                         if providers.contains(&provider) {
+                            if provider
+                                == rbw::api::TwoFactorProviderType::Email
+                            {
+                                if let Some(sso_email_2fa_session_token) =
+                                    sso_email_2fa_session_token
+                                {
+                                    rbw::actions::send_two_factor_email(
+                                        &email,
+                                        &sso_email_2fa_session_token,
+                                    )
+                                    .await?;
+                                }
+                            }
                             let (
                                 access_token,
                                 refresh_token,
@@ -161,7 +175,7 @@ pub async fn login(
                                 parallelism,
                                 protected_key,
                             ) = two_factor(
-                                tty,
+                                environment,
                                 &email,
                                 password.clone(),
                                 provider,
@@ -184,7 +198,9 @@ pub async fn login(
                             break 'attempts;
                         }
                     }
-                    return Err(anyhow::anyhow!("TODO"));
+                    return Err(anyhow::anyhow!(
+                        "unsupported two factor methods: {providers:?}"
+                    ));
                 }
                 Err(rbw::error::Error::IncorrectPassword { message }) => {
                     if i == 3 {
@@ -194,7 +210,6 @@ pub async fn login(
                         .context("failed to log in to bitwarden instance");
                     }
                     err_msg = Some(message);
-                    continue;
                 }
                 Err(e) => {
                     return Err(e)
@@ -210,7 +225,7 @@ pub async fn login(
 }
 
 async fn two_factor(
-    tty: Option<&str>,
+    environment: &rbw::protocol::Environment,
     email: &str,
     password: rbw::locked::Password,
     provider: rbw::api::TwoFactorProviderType,
@@ -237,7 +252,7 @@ async fn two_factor(
             provider.header(),
             provider.message(),
             err.as_deref(),
-            tty,
+            environment,
             provider.grab(),
         )
         .await
@@ -279,7 +294,6 @@ async fn two_factor(
                     .context("failed to log in to bitwarden instance");
                 }
                 err_msg = Some(message);
-                continue;
             }
             // can get this if the user passes an empty string
             Err(rbw::error::Error::TwoFactorRequired { .. }) => {
@@ -291,7 +305,6 @@ async fn two_factor(
                     .context("failed to log in to bitwarden instance");
                 }
                 err_msg = Some(message);
-                continue;
             }
             Err(e) => {
                 return Err(e)
@@ -304,7 +317,7 @@ async fn two_factor(
 }
 
 async fn login_success(
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
     access_token: String,
     refresh_token: String,
     kdf: rbw::api::KdfType,
@@ -316,13 +329,13 @@ async fn login_success(
     mut db: rbw::db::Db,
     email: String,
 ) -> anyhow::Result<()> {
-    db.access_token = Some(access_token.to_string());
-    db.refresh_token = Some(refresh_token.to_string());
+    db.access_token = Some(access_token.clone());
+    db.refresh_token = Some(refresh_token.clone());
     db.kdf = Some(kdf);
     db.iterations = Some(iterations);
     db.memory = memory;
     db.parallelism = parallelism;
-    db.protected_key = Some(protected_key.to_string());
+    db.protected_key = Some(protected_key.clone());
     save_db(&db).await?;
 
     sync(None, state.clone()).await?;
@@ -358,10 +371,9 @@ async fn login_success(
     Ok(())
 }
 
-pub async fn unlock(
-    sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
-    tty: Option<&str>,
+async fn unlock_state(
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    environment: &rbw::protocol::Environment,
 ) -> anyhow::Result<()> {
     if state.lock().await.needs_unlock() {
         let db = load_db().await?;
@@ -409,7 +421,7 @@ pub async fn unlock(
                     rbw::dirs::profile()
                 ),
                 err.as_deref(),
-                tty,
+                environment,
                 true,
             )
             .await
@@ -437,20 +449,29 @@ pub async fn unlock(
                         .context("failed to unlock database");
                     }
                     err_msg = Some(message);
-                    continue;
                 }
                 Err(e) => return Err(e).context("failed to unlock database"),
             }
         }
     }
 
+    Ok(())
+}
+
+pub async fn unlock(
+    sock: &mut crate::sock::Sock,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    environment: &rbw::protocol::Environment,
+) -> anyhow::Result<()> {
+    unlock_state(state, environment).await?;
+
     respond_ack(sock).await?;
 
     Ok(())
 }
 
 async fn unlock_success(
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
     keys: rbw::locked::Keys,
     org_keys: std::collections::HashMap<String, rbw::locked::Keys>,
 ) -> anyhow::Result<()> {
@@ -462,7 +483,7 @@ async fn unlock_success(
 
 pub async fn lock(
     sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
 ) -> anyhow::Result<()> {
     state.lock().await.clear();
 
@@ -473,7 +494,7 @@ pub async fn lock(
 
 pub async fn check_lock(
     sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
 ) -> anyhow::Result<()> {
     if state.lock().await.needs_unlock() {
         return Err(anyhow::anyhow!("agent is locked"));
@@ -486,7 +507,7 @@ pub async fn check_lock(
 
 pub async fn sync(
     sock: Option<&mut crate::sock::Sock>,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
 ) -> anyhow::Result<()> {
     let mut db = load_db().await?;
 
@@ -506,6 +527,7 @@ pub async fn sync(
     ) = rbw::actions::sync(&access_token, &refresh_token)
         .await
         .context("failed to sync database from server")?;
+    state.lock().await.set_master_password_reprompt(&entries);
     if let Some(access_token) = access_token {
         db.access_token = Some(access_token);
     }
@@ -526,27 +548,141 @@ pub async fn sync(
     Ok(())
 }
 
-pub async fn decrypt(
-    sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+async fn decrypt_cipher(
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    environment: &rbw::protocol::Environment,
     cipherstring: &str,
+    entry_key: Option<&str>,
     org_id: Option<&str>,
-) -> anyhow::Result<()> {
-    let state = state.lock().await;
+) -> anyhow::Result<String> {
+    let mut state = state.lock().await;
+    if !state.master_password_reprompt_initialized() {
+        let db = load_db().await?;
+        state.set_master_password_reprompt(&db.entries);
+    }
     let Some(keys) = state.key(org_id) else {
         return Err(anyhow::anyhow!(
             "failed to find decryption keys in in-memory state"
         ));
     };
+    let entry_key = if let Some(entry_key) = entry_key {
+        let key_cipherstring =
+            rbw::cipherstring::CipherString::new(entry_key)
+                .context("failed to parse individual item encryption key")?;
+        Some(rbw::locked::Keys::new(
+            key_cipherstring.decrypt_locked_symmetric(keys).context(
+                "failed to decrypt individual item encryption key",
+            )?,
+        ))
+    } else {
+        None
+    };
+
+    let mut sha256 = sha2::Sha256::new();
+    sha256.update(cipherstring);
+    let master_password_reprompt: [u8; 32] = sha256.finalize().into();
+    if state
+        .master_password_reprompt
+        .contains(&master_password_reprompt)
+    {
+        let db = load_db().await?;
+
+        let Some(kdf) = db.kdf else {
+            return Err(anyhow::anyhow!("failed to find kdf type in db"));
+        };
+
+        let Some(iterations) = db.iterations else {
+            return Err(anyhow::anyhow!(
+                "failed to find number of iterations in db"
+            ));
+        };
+
+        let memory = db.memory;
+        let parallelism = db.parallelism;
+
+        let Some(protected_key) = db.protected_key else {
+            return Err(anyhow::anyhow!(
+                "failed to find protected key in db"
+            ));
+        };
+        let Some(protected_private_key) = db.protected_private_key else {
+            return Err(anyhow::anyhow!(
+                "failed to find protected private key in db"
+            ));
+        };
+
+        let email = config_email().await?;
+
+        let mut err_msg = None;
+        for i in 1_u8..=3 {
+            let err = if i > 1 {
+                // this unwrap is safe because we only ever continue the loop
+                // if we have set err_msg
+                Some(format!("{} (attempt {}/3)", err_msg.unwrap(), i))
+            } else {
+                None
+            };
+            let password = rbw::pinentry::getpin(
+                &config_pinentry().await?,
+                "Master Password",
+                "Accessing this entry requires the master password",
+                err.as_deref(),
+                environment,
+                true,
+            )
+            .await
+            .context("failed to read password from pinentry")?;
+            match rbw::actions::unlock(
+                &email,
+                &password,
+                kdf,
+                iterations,
+                memory,
+                parallelism,
+                &protected_key,
+                &protected_private_key,
+                &db.protected_org_keys,
+            ) {
+                Ok(_) => {
+                    break;
+                }
+                Err(rbw::error::Error::IncorrectPassword { message }) => {
+                    if i == 3 {
+                        return Err(rbw::error::Error::IncorrectPassword {
+                            message,
+                        })
+                        .context("failed to unlock database");
+                    }
+                    err_msg = Some(message);
+                }
+                Err(e) => return Err(e).context("failed to unlock database"),
+            }
+        }
+    }
+
     let cipherstring = rbw::cipherstring::CipherString::new(cipherstring)
         .context("failed to parse encrypted secret")?;
     let plaintext = String::from_utf8(
         cipherstring
-            .decrypt_symmetric(keys)
+            .decrypt_symmetric(keys, entry_key.as_ref())
             .context("failed to decrypt encrypted secret")?,
     )
     .context("failed to parse decrypted secret")?;
 
+    Ok(plaintext)
+}
+
+pub async fn decrypt(
+    sock: &mut crate::sock::Sock,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    environment: &rbw::protocol::Environment,
+    cipherstring: &str,
+    entry_key: Option<&str>,
+    org_id: Option<&str>,
+) -> anyhow::Result<()> {
+    let plaintext =
+        decrypt_cipher(state, environment, cipherstring, entry_key, org_id)
+            .await?;
     respond_decrypt(sock, plaintext).await?;
 
     Ok(())
@@ -554,7 +690,7 @@ pub async fn decrypt(
 
 pub async fn encrypt(
     sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
     plaintext: &str,
     org_id: Option<&str>,
 ) -> anyhow::Result<()> {
@@ -575,28 +711,41 @@ pub async fn encrypt(
     Ok(())
 }
 
+#[cfg(feature = "clipboard")]
 pub async fn clipboard_store(
     sock: &mut crate::sock::Sock,
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
     text: &str,
 ) -> anyhow::Result<()> {
-    state
-        .lock()
-        .await
-        .clipboard
-        .set_contents(text.to_owned())
-        .map_err(|e| {
+    let mut state = state.lock().await;
+    if let Some(clipboard) = &mut state.clipboard {
+        clipboard.set_text(text).map_err(|e| {
             anyhow::anyhow!("couldn't store value to clipboard: {e}")
         })?;
+    }
 
     respond_ack(sock).await?;
 
     Ok(())
 }
 
+#[cfg(not(feature = "clipboard"))]
+pub async fn clipboard_store(
+    sock: &mut crate::sock::Sock,
+    _state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    _text: &str,
+) -> anyhow::Result<()> {
+    sock.send(&rbw::protocol::Response::Error {
+        error: "clipboard not supported".to_string(),
+    })
+    .await?;
+
+    Ok(())
+}
+
 pub async fn version(sock: &mut crate::sock::Sock) -> anyhow::Result<()> {
     sock.send(&rbw::protocol::Response::Version {
-        version: rbw::protocol::version(),
+        version: rbw::protocol::VERSION,
     })
     .await?;
 
@@ -670,7 +819,7 @@ async fn config_pinentry() -> anyhow::Result<String> {
 }
 
 pub async fn subscribe_to_notifications(
-    state: std::sync::Arc<tokio::sync::Mutex<crate::agent::State>>,
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
 ) -> anyhow::Result<()> {
     if state.lock().await.notifications_handler.is_connected() {
         return Ok(());
@@ -700,3 +849,104 @@ pub async fn subscribe_to_notifications(
         .err()
         .map_or_else(|| Ok(()), |err| Err(anyhow::anyhow!(err.to_string())))
 }
+
+pub async fn get_ssh_public_keys(
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+) -> anyhow::Result<Vec<String>> {
+    let environment = {
+        let state = state.lock().await;
+        state.set_timeout();
+        state.last_environment().clone()
+    };
+    unlock_state(state.clone(), &environment).await?;
+
+    let db = load_db().await?;
+    let mut pubkeys = Vec::new();
+
+    for entry in db.entries {
+        if let rbw::db::EntryData::SshKey {
+            public_key: Some(encrypted),
+            ..
+        } = &entry.data
+        {
+            let plaintext = decrypt_cipher(
+                state.clone(),
+                &environment,
+                encrypted,
+                entry.key.as_deref(),
+                entry.org_id.as_deref(),
+            )
+            .await?;
+
+            pubkeys.push(plaintext);
+        }
+    }
+
+    Ok(pubkeys)
+}
+
+pub async fn find_ssh_private_key(
+    state: std::sync::Arc<tokio::sync::Mutex<crate::state::State>>,
+    request_public_key: ssh_agent_lib::ssh_key::PublicKey,
+) -> anyhow::Result<ssh_agent_lib::ssh_key::PrivateKey> {
+    let environment = {
+        let state = state.lock().await;
+        state.set_timeout();
+        state.last_environment().clone()
+    };
+    unlock_state(state.clone(), &environment).await?;
+
+    let request_bytes = request_public_key.to_bytes();
+
+    let db = load_db().await?;
+
+    for entry in db.entries {
+        if let rbw::db::EntryData::SshKey {
+            private_key,
+            public_key,
+            ..
+        } = &entry.data
+        {
+            let Some(public_key_enc) = public_key else {
+                continue;
+            };
+            let public_key_plaintext = decrypt_cipher(
+                state.clone(),
+                &environment,
+                public_key_enc,
+                entry.key.as_deref(),
+                entry.org_id.as_deref(),
+            )
+            .await?;
+            let public_key_bytes =
+                ssh_agent_lib::ssh_key::PublicKey::from_openssh(
+                    &public_key_plaintext,
+                )
+                .map_err(anyhow::Error::new)?
+                .to_bytes();
+
+            if public_key_bytes == request_bytes {
+                let private_key_enc =
+                    private_key.as_ref().ok_or_else(|| {
+                        anyhow::anyhow!("Matching entry has no private key")
+                    })?;
+
+                let private_key_plaintext = decrypt_cipher(
+                    state.clone(),
+                    &environment,
+                    private_key_enc,
+                    entry.key.as_deref(),
+                    entry.org_id.as_deref(),
+                )
+                .await?;
+
+                return ssh_agent_lib::ssh_key::PrivateKey::from_openssh(
+                    private_key_plaintext,
+                )
+                .map_err(anyhow::Error::new);
+            }
+        }
+    }
+
+    Err(anyhow::anyhow!("No matching private key found"))
+}
